/**********************************************************************
** This program is part of 'MOOSE', the
** Messaging Object Oriented Simulation Environment.
**           Copyright (C) 2003-2013 Upinder S. Bhalla. and NCBS
** It is made available under the terms of the
** GNU Lesser General Public License version 2.1
** See the file COPYING.LIB for the full notice.
**********************************************************************/


#include "../basecode/header.h"
#include "../shell/Shell.h"
#include "../biophysics/CompartmentBase.h"
#include "../biophysics/Compartment.h"
#include "IntFireBase.h"

// This tests stuff without using the messaging.
void testIntFire()
{
#if 0
 	test network
#endif
}

// This is applicable to tests that use the messaging and scheduling.
void testIntFireProcess()
{
}
