/***
 *    Description:  Tests functions in global.h
 *
 *        Created:  2020-04-03

 *         Author:  Dilawar Singh <dilawar.s.rajput@gmail.com>
 *        License:  MIT License
 */

#include "global.h"
#include "../utility/simple_assert.hpp"


int main(int argc, const char *argv[])
{
    // Add tests here.
    //
    return 0;
}

