//
//  Copyright Toon Knapen
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_NUMERIC_BINDINGS_AMOS_AMOS_NAMES_H
#define BOOST_NUMERIC_BINDINGS_AMOS_AMOS_NAMES_H

#include <boost/numeric/bindings/traits/fortran.h>

#define AMOS_DBESI FORTRAN_ID( dbesi )
#define AMOS_CBESI FORTRAN_ID( cbesi )
#define AMOS_ZBESI FORTRAN_ID( zbesi )

#define AMOS_DBESJ FORTRAN_ID( dbesj )
#define AMOS_CBESJ FORTRAN_ID( cbesj )
#define AMOS_ZBESJ FORTRAN_ID( zbesj )

#define AMOS_DBESY FORTRAN_ID( dbesy )
#define AMOS_CBESY FORTRAN_ID( cbesy )
#define AMOS_ZBESY FORTRAN_ID( zbesy )

#endif // BOOST_NUMERIC_BINDINGS_AMOS_AMOS_NAMES_HPP
