//
// Copyright Kresimir Fresl, Toon Knapen, and Karl Meerbergen 2002, 2003
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_BINDINGS_TRANSPOSE_HPP
#define BOOST_BINDINGS_TRANSPOSE_HPP

namespace boost { namespace numeric { namespace bindings { namespace traits {

  const char NO_TRANSPOSE = 'N' ;
  const char TRANSPOSE    = 'T' ;
  const char CONJUGATE    = 'C' ;

}}}}

#endif
