/*
 * ==============================================================================
 *
 *       Filename:  simpletest.hpp
 *
 *    Description:  Main header file containing all definitions and macros.
 *
 *        Version:  1.0
 *        Created:  05/20/2014 09:22:42 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Dilawar Singh (), dilawars@ncbs.res.in
 *   Organization:
 *
 * ==============================================================================
 */

#ifndef  SIMPLETEST_INC
#define  SIMPLETEST_INC

#include "print_function.hpp"
#include "testing_macros.hpp"
#include "simple_assert.hpp"

#endif   /* ----- #ifndef SIMPLETEST_INC  ----- */
